// MainFrm.cpp : implementation of the CMainFrame class
//

#include "stdafx.h"
#include "Stage.h"
#include "MainFrm.h"
#include "indevdlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CMainFrame

IMPLEMENT_DYNCREATE(CMainFrame, CFrameWnd)

BEGIN_MESSAGE_MAP(CMainFrame, CFrameWnd)
	//{{AFX_MSG_MAP(CMainFrame)
	ON_WM_CREATE()
	ON_WM_DESTROY()
	ON_WM_ACTIVATEAPP()
	ON_WM_PALETTECHANGED()
	ON_WM_MOVE()
	ON_COMMAND(ID_FILE_NEW, OnFileNew)
	ON_COMMAND(ID_FILE_OPEN, OnFileOpen)
	ON_COMMAND(ID_EDIT_INSERTSHAPE, OnEditInsertShape)
	ON_WM_SETFOCUS()
	ON_COMMAND(ID_EDIT_CONTROL, OnEditControl)
	ON_UPDATE_COMMAND_UI(ID_EDIT_CONTROL, OnUpdateEditControl)
	ON_COMMAND(ID_EDIT_INPUTDEV, OnEditInputdev)
	ON_COMMAND(ID_EDIT_INSHEAD, OnEditInshead)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

static UINT indicators[] =
{
	ID_SEPARATOR,           // status line indicator
	ID_INDICATOR_CAPS,
	ID_INDICATOR_NUM,
	ID_INDICATOR_SCRL,
};

/////////////////////////////////////////////////////////////////////////////
// CMainFrame construction/destruction

CMainFrame::CMainFrame()
{
	m_pScene = NULL;
	m_pCurShape = FALSE;
	m_pController = NULL;
	m_iObjCtrlDev = 0;
	m_iControlObj = 0;
	m_iControlType = 0;
	m_pDirLight = NULL;
	m_bPhase = FALSE;
}

CMainFrame::~CMainFrame()
{
}

int CMainFrame::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	if (CFrameWnd::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	if (!m_wndToolBar.Create(this) ||
		!m_wndToolBar.LoadToolBar(IDR_MAINFRAME))
	{
		TRACE0("Failed to create toolbar\n");
		return -1;      // fail to create
	}

	if (!m_wndStatusBar.Create(this) ||
		!m_wndStatusBar.SetIndicators(indicators,
		  sizeof(indicators)/sizeof(UINT)))
	{
		TRACE0("Failed to create status bar\n");
		return -1;      // fail to create
	}

	// Create the 3D window
	if (!m_wnd3d.Create(this, IDC_3DWND, D3DCOLOR_RGB)) {
		return -1;
	}

	// TODO: Remove this if you don't want tool tips or a resizeable toolbar
	m_wndToolBar.SetBarStyle(m_wndToolBar.GetBarStyle() |
		CBRS_TOOLTIPS | CBRS_FLYBY | CBRS_SIZE_DYNAMIC);

	// TODO: Delete these three lines if you don't want the toolbar to
	//  be dockable
	m_wndToolBar.EnableDocking(CBRS_ALIGN_ANY);
	EnableDocking(CBRS_ALIGN_ANY);
	DockControlBar(&m_wndToolBar);

	NewScene();
	ASSERT(m_pScene);

	// Create the controller object
	if (m_iControlType == 0) {
		m_pController = new C3dPosCtlr;
	} else {
		m_pController = new C3dFlyCtlr;
	}
	m_pController->Create(&m_wnd3d,
						  OnGetCtrlFrame,
					      this);

	// restore the controller settings
	m_pController->SelectDevice(m_iObjCtrlDev);

	// enable mouse selection of objects
	m_wnd3d.EnableMouseSelection(OnSelChange, this);

	// Insert the head
	OnEditInshead();

	return 0;
}

BOOL CMainFrame::PreCreateWindow(CREATESTRUCT& cs)
{
	// Restore any saved settings
	RestoreSettings();

	// Restore the window size and position
	cs.x = m_rcWnd.left;
	cs.y = m_rcWnd.top;
	cs.cx = m_rcWnd.Width();
	cs.cy = m_rcWnd.Height();

	return CFrameWnd::PreCreateWindow(cs);
}

BOOL CMainFrame::NewScene()
{

	// Delete any scene we might have
	if (m_pScene) {
		m_wnd3d.SetScene(NULL);
		delete m_pScene;
		m_pScene = NULL;
	}
	MakeCurrent(NULL);
	if (m_pDirLight) delete m_pDirLight;

	// Create an initial scene
	m_pScene = new C3dScene;
	if (!m_pScene->Create()) return FALSE;

	// Set up the lighting
    m_pScene->SetAmbientLight(0.2, 0.2, 0.2);
	m_pDirLight = new C3dDirLight;
	m_pDirLight->Create(1, 0, 0);
    m_pDirLight->SetPosition(-2, 2, -5);
	m_pDirLight->SetDirection(1, -1, 1);
    m_pScene->AddChild(m_pDirLight);

	// reset the camera position and direction
    m_pScene->SetCameraPosition(C3dVector(0, 0, -10));
    m_pScene->SetCameraDirection(C3dVector(0, 0, 1), C3dVector(0, 1, 0));


	m_wnd3d.SetScene(m_pScene);
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CMainFrame diagnostics

#ifdef _DEBUG
void CMainFrame::AssertValid() const
{
	CFrameWnd::AssertValid();
}

void CMainFrame::Dump(CDumpContext& dc) const
{
	CFrameWnd::Dump(dc);
}

#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CMainFrame message handlers

void CMainFrame::RecalcLayout(BOOL bNotify) 
{
    // Rearrange the control bars and fit the 3D window in the middle
	// Let the frame rearrange the control bars
	CFrameWnd::RecalcLayout(bNotify);

	// Find the space that's left over
	CRect rc;
	RepositionBars(0,
				   0xFFFF,
				   IDC_3DWND,
				   CWnd::reposQuery,
				   &rc);
	if (IsWindow(m_wnd3d.GetSafeHwnd())) {
		m_wnd3d.MoveWindow(&rc, FALSE);
	}
}


void CMainFrame::OnDestroy() 
{
	CFrameWnd::OnDestroy();

	// Save the current settings
	SaveSettings();

	if (m_pDirLight) delete m_pDirLight;
	
	// destroy the controller
	if (m_pController) {
		delete m_pController;
		m_pController = NULL;
	}

	// Delete any scene we might have
	m_wnd3d.SetScene(NULL);
	if (m_pScene) {
		delete m_pScene;
		m_pScene = NULL;
	}
	MakeCurrent(NULL);
}

void CMainFrame::OnActivateApp(BOOL bActive, HTASK hTask) 
{
	CFrameWnd::OnActivateApp(bActive, hTask);
	
	// Tell the 3D window about the new state
	m_wnd3d.SendMessage(WM_ACTIVATEAPP,
						(WPARAM)bActive,
						(LPARAM)hTask);
}

void CMainFrame::OnPaletteChanged(CWnd* pFocusWnd) 
{
	// Let the 3D window know
	m_wnd3d.SendMessage(WM_PALETTECHANGED,
					   pFocusWnd ? (WPARAM)pFocusWnd->GetSafeHwnd() : 0);
}

void CMainFrame::OnMove(int x, int y) 
{
	CFrameWnd::OnMove(x, y);
	
	// Let the 3D window know the frame has moved
	m_wnd3d.SendMessage(WM_MOVE, 0, MAKELPARAM(0, 0));
}

void CMainFrame::OnFileNew() 
{
	NewScene();
}

BOOL CMainFrame::OpenFile(const char* pszPath)
{
	// Create a new scene
	NewScene();
	ASSERT(m_pScene);

	// Try for a shape file
	m_pCurShape = new C3dShape;
	const char* pszFile = m_pCurShape->Load(pszPath);
	if (!pszFile) return FALSE;

	// Attach the new shape to the scene
	m_pScene->AddChild(m_pCurShape);

	// Add the name to the recent file list
	AfxGetApp()->AddToRecentFileList(pszFile);
	Status("Opened: %s", pszFile);
	return TRUE;
}

void CMainFrame::OnFileOpen() 
{
	OpenFile(NULL);
}

void CMainFrame::OnEditInsertShape() 
{
	CShapeDlg dlg("Insert Shape", this);

	if (dlg.DoModal() != IDOK) return;

	// Add the new shape
	if (dlg.m_pShape) {
		ASSERT(m_pScene);
		m_pScene->AddChild(dlg.m_pShape);

		if (m_pCurShape) delete m_pCurShape;
		m_pCurShape = dlg.m_pShape;
	}
}

void CMainFrame::Status(const char* pszFormat, ...)
{
	char buf[256];
	va_list arglist;
	va_start(arglist, pszFormat);
    vsprintf(buf, pszFormat, arglist);
	va_end(arglist);
	m_wndStatusBar.SetWindowText(buf);
	m_wndStatusBar.UpdateWindow();
}

void CMainFrame::OnSetFocus(CWnd* pOldWnd) 
{
	CFrameWnd::OnSetFocus(pOldWnd);
	
	// Give the focus to the 3D window so it can process
	// keystrokes for the key input device
	m_wnd3d.SetFocus();
}

//////////////////////////////////////////////////////////////
// Control device selection

// callback from the dialog to get a frame to control
// static
C3dFrame* CMainFrame::OnGetCtrlFrame(void* pArg)
{
	CMainFrame* pThis = (CMainFrame*) pArg;
	ASSERT(pThis);
	ASSERT(pThis->IsKindOf(RUNTIME_CLASS(CMainFrame)));
	if (pThis->m_iControlObj == 1) {
		return pThis->m_wnd3d.GetCamera();
	}
	return pThis->m_pCurShape;
}

void CMainFrame::OnEditControl() 
{
	CInpDevDlg dlg;
	dlg.m_iObject = m_iControlObj;
	dlg.m_iType = m_iControlType;
	dlg.m_iDevice = m_iObjCtrlDev;
	if (dlg.DoModal() != IDOK) return;
	m_iControlObj = dlg.m_iObject;

	// set up the controller type
	if (dlg.m_iType != m_iControlType) {

		if (m_pController) delete m_pController;

		// create a new controller
		m_iControlType = dlg.m_iType;
		if (m_iControlType == 0) {
			m_pController = new C3dPosCtlr;
		} else {
			m_pController = new C3dFlyCtlr;
		}
		m_pController->Create(&m_wnd3d,
							  OnGetCtrlFrame,
							  this);
	}

	ASSERT(m_pController);
	m_pController->SelectDevice(dlg.m_iDevice);
 	m_iObjCtrlDev = m_pController->GetCurDevice();
	Status("Control is now via: %s", 
		   m_pController->GetDeviceName(m_iObjCtrlDev));
}

void CMainFrame::OnUpdateEditControl(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_pController ? TRUE : FALSE);
}

void CMainFrame::OnEditInputdev() 
{
	ASSERT(m_pController);
	C3dInputDevice* pDev = m_pController->GetInputDevice();
	ASSERT(pDev);
	pDev->ShowConfigDlg();
}

//////////////////////////////////////////////////////////////////
// Save and Restore Settings

void CMainFrame::SaveSettings()
{
	CWinApp* pa = AfxGetApp();
	pa->WriteProfileInt("Settings","ControlDevice", m_iObjCtrlDev);
	CRect rc;
	GetWindowRect(&rc);
	pa->WriteProfileInt("Settings\\Window","Top", rc.top);
	pa->WriteProfileInt("Settings\\Window","Left", rc.left);
	pa->WriteProfileInt("Settings\\Window","Width", rc.Width());
	pa->WriteProfileInt("Settings\\Window","Height", rc.Height());
}

void CMainFrame::RestoreSettings()
{
	CWinApp* pa = AfxGetApp();
	m_iObjCtrlDev = pa->GetProfileInt("Settings","ControlDevice", 0);
	m_rcWnd.top = pa->GetProfileInt("Settings\\Window","Top", 50);
	m_rcWnd.left = pa->GetProfileInt("Settings\\Window","Left", 50);
	m_rcWnd.right = m_rcWnd.left + pa->GetProfileInt("Settings\\Window","Width", 300); 
	m_rcWnd.bottom = m_rcWnd.top + pa->GetProfileInt("Settings\\Window","Height", 350); 
}

//////////////////////////////////////////////////////////////////////////
// object selection

// callback notification function
// called by m_wnd3d when the object selection changes
// static
void CMainFrame::OnSelChange(C3dShape* pShape, CPoint pt, void* pArg)
{
	// Get a pointer to the class
	CMainFrame* pThis = (CMainFrame*) pArg;
	ASSERT(pThis);
	ASSERT(pThis->IsKindOf(RUNTIME_CLASS(CMainFrame)));

	// Make the selected shape the current one
	pThis->MakeCurrent(pShape);
}

void CMainFrame::MakeCurrent(C3dShape* pShape)
{
	m_pCurShape = pShape;
	if (m_pCurShape != NULL) {
		Status("Selected: %s", m_pCurShape->GetName());
	} else {
		Status("No selection");
	}
}

void CMainFrame::OnEditInshead() 
{
	// Insert a head
	C3dShape* pShape = new C3dShape;
	pShape->Load(IDX_HEAD);
	m_pScene->AddChild(pShape);
	m_pScene->m_ShapeList.Append(pShape);
	pShape->SetPosition(0, 0, 3);
	pShape->SetName("Head");
	MakeCurrent(pShape);
}

BOOL CMainFrame::Update()
{
	BOOL bMore = FALSE;

	// See if the object controller wants to do anything
	if (m_pController) {
		m_pController->Update();
	}

	// Do the phased lighting thing
	m_bPhase = !m_bPhase;
	double d = 1.0;	// Separation
	double cz = 10; // Camera z position
	C3dVector vo;
	if (m_pCurShape) {
		m_pCurShape->GetPosition(vo);
	} else {
		vo = C3dVector(0, 0, 0);
	}
	C3dVector cv;
	if (m_pDirLight && m_pScene) {
		if (m_bPhase) {
			m_pDirLight->SetColor(1, 0, 0);
			cv = C3dVector(-d, 0, -cz);
		} else {
			m_pDirLight->SetColor(0, 1, 0);
			cv = C3dVector(d, 0, -cz);
		}
		m_pScene->SetCameraPosition(cv);
		C3dVector vl = vo - cv;
		m_pScene->SetCameraDirection(vl);
	}

	// Update 3D window
	if (m_wnd3d.Update(1)) {
		bMore = TRUE;
	}
	
	return bMore;
}